###############################################################################
# (c) Copyright 2020-2024 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import asyncio
import glob
import os
import pathlib
import re
import subprocess
import tempfile
from datetime import datetime
from os.path import dirname, join

import apd
import click
from LbAPCommon import write_jsroot_compression_options
from LbAPCommon.hacks import setup_lbrun_environment

LBAPI_HOST = os.environ.get("LBAPI_HOST", "https://lbap.app.cern.ch")

RE_PRODUCTION_NAME = re.compile(
    r"^(?:[a-z][a-z0-9_]+/)*[a-z][a-z0-9_]+/?$", re.IGNORECASE
)


def validate_environment():
    click.secho("Validating environment", fg="green")
    check_proxy()
    check_cvmfs()


def inside_ap_datapkg():
    """Check if script is run from main directory"""
    from pathlib import Path

    cwd = Path.cwd()
    xenv_file = cwd / "AnalysisProductions.xenv"

    if not xenv_file.exists():
        raise click.ClickException(
            "Running command in wrong directory! Please run from the "
            "AnalysisProductions base folder or use the `--branch` / `-C` flags.\n"
            f"Current directory: {cwd}\n"
            f"Looking for: {xenv_file}"
        )
    os.environ["APPTAINER_BINDPATH"] = str(cwd)


def check_proxy():
    try:
        subprocess.check_call(
            ["lb-dirac", "dirac-proxy-info", "--checkvalid"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        raise click.ClickException(
            "No grid proxy found, please get one with lhcb-proxy-init"
        )


def check_cvmfs():
    for path in ["/cvmfs/lhcb.cern.ch", "/cvmfs/lhcb-condb.cern.ch"]:
        if not os.listdir(path):
            raise click.ClickException(f"Missing CVMFS repository: {path}")


def pool_xml_catalog(lfns):
    click.secho("Generating pool XML catalog", fg="green")

    # First check if we already have access
    with tempfile.TemporaryDirectory() as tmp_dir:
        proc = subprocess.run(
            ["lb-dirac", "dirac-bookkeeping-genXMLCatalog", "--LFNs", ",".join(lfns)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=tmp_dir,
        )
        if proc.returncode != 0:
            click.secho("********** stdout was:", fg="red")
            click.echo(proc.stdout)
            click.secho("********** stderr was:", fg="red")
            click.echo(proc.stderr)
            raise click.ClickException(
                f"Failed to generate pool XML catalog with: {proc.args}"
            )
        with open(join(tmp_dir, "pool_xml_catalog.xml"), "rt") as fp:
            return fp.read()


def available_productions():
    """Function that finds all production folders with an info.yaml"""

    # Glob all the info.yaml and extract production name
    folders_with_info = glob.glob("*/info.yaml", recursive=True)
    production_names = []
    for folder in folders_with_info:
        folder_name = dirname(folder)
        production_names.append(folder_name)

    return production_names


def check_production(production_name):
    if production_name not in available_productions():
        raise click.ClickException(
            f"Can't find production {production_name}. Does it have an info.yaml?"
        )
    if not RE_PRODUCTION_NAME.match(production_name):
        raise click.ClickException(
            f"Production name must match the pattern {RE_PRODUCTION_NAME.pattern!r}. "
            "Please choose a different name."
        )


def check_production_existed(available_versions: dict, production_name: str):
    for productions in list(available_versions.values()):
        if production_name in productions:
            return True
    raise click.ClickException(
        f"Can't find production {production_name} for any version tag of AnalysisProductions."
    )


def create_output_dir(production_name, require_cmt):
    """Create the directory structure for testing locally"""
    date_string = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    testing_dir = (
        pathlib.Path.cwd() / "local-tests" / f"{production_name}-{date_string}"
    )
    click.secho(f"Running tests in {testing_dir}", fg="green")
    testing_dir.mkdir(parents=True)

    dynamic_dir = testing_dir / "dynamic"
    dynamic_dir.mkdir(parents=True)
    out_dir = testing_dir / "output"
    out_dir.mkdir(parents=True)

    main_dynamic_dir = pathlib.Path.cwd() / "dynamic"
    if main_dynamic_dir.exists():
        click.secho(
            f"Found existing dynamic dir pointing to "
            f"{main_dynamic_dir.resolve()}, unlinking",
            fg="green",
        )
        main_dynamic_dir.unlink()
    main_dynamic_dir.symlink_to(dynamic_dir)
    write_jsroot_compression_options(dynamic_dir)

    # Create the fake install directory for lb-run
    click.secho(f"Setting CMAKE_PREFIX_PATH to {testing_dir}", fg="green")
    setup_lbrun_environment(testing_dir, os.getcwd(), require_cmt)

    return dynamic_dir, out_dir


def logging_subprocess_run(args, *, cwd=None, printing=True):
    return asyncio.run(_logging_subprocess_run(*args, cwd=cwd, printing=printing))


async def _logging_subprocess_run(*command, cwd, printing):
    proc = await asyncio.create_subprocess_exec(
        *command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd,
        # Increase the maximum possible line length to 10MB
        limit=10 * 1024 * 1024,
    )
    stdout, stderr, returncode = await asyncio.gather(
        _handle_output(proc.stdout, printing),
        _handle_output(proc.stderr, printing),
        proc.wait(),
    )
    return subprocess.CompletedProcess(command, returncode, stdout, stderr)


async def _handle_output(stream, printing: bool):
    """Process the output of a current local execution"""
    result = b""
    while line := await stream.readline():
        result += line
        if printing:
            click.echo(line.decode(errors="backslashreplace"), nl=False)
    return result


def production_name_type(production_name: str) -> str:
    return production_name.strip("/")


def recursively_create_input(production_name, job_name, filetype, prod_data, n_events):
    click.secho(
        f"Running {job_name} to provide the necessary input",
        fg="green",
    )

    output_filetypes = [x.lower() for x in prod_data[job_name]["output"]]
    if filetype:
        if filetype.lower() not in output_filetypes:
            raise NotImplementedError(
                f"Looking for {filetype} but {job_name} only produces {output_filetypes}"
            )
    else:
        if len(output_filetypes) > 1:
            raise ValueError(
                "An input filetype must be specified when taking input from a multi-output job!"
            )
        else:
            filetype = output_filetypes[0]

    # Run the job the desired job depends on
    run_args = ["lb-ap-legacy", "test"]
    if n_events is not None:
        run_args += ["--n-events", str(n_events)]
    run_args += [production_name, job_name]
    result = logging_subprocess_run(run_args).stdout.decode()
    # Get the output directory
    output_dir = result.split("\n")[-2].split(" ")[-1]
    # if this is False clearly some other error has occurred so don't bother continuing
    if os.path.exists(output_dir):
        dependent_input = output_dir + f"/00012345_00006789_1.{filetype}"
        for file_path in glob.glob(f"{output_dir}/*"):
            if re.match(
                dependent_input,
                file_path,
                re.IGNORECASE,
            ):
                dependent_input = file_path
                break
        else:
            raise OSError(
                f"{job_name} did not produce the desired output {dependent_input}"
            )
        other_files = list(pathlib.Path(output_dir).glob("*.tck.json"))

        return dependent_input, other_files
    else:
        raise OSError(f"The expected output directory {output_dir} was not found! ")


def resolve_input_overrides(prod_data, job_name):
    override_input = None
    output_filetype = None
    if dependent_job := prod_data[job_name]["input"].get("job_name"):
        job_data = prod_data[dependent_job]

        if "bk_query" in job_data["input"]:
            override_input = job_data["input"]
            output_filetype = job_data["output"][0]
        elif "transform_ids" in job_data["input"]:
            override_input = job_data["input"]
            output_filetype = job_data["output"][0]
        elif "job_name" in job_data["input"]:
            override_input, output_filetype = resolve_input_overrides(
                prod_data, job_data["input"]["job_name"]
            )
        else:
            raise NotImplementedError(
                "Input requires either a bookkeeping location, transform ids or a previous job name"
            )
    return override_input, output_filetype


def check_status_code(response):
    if not response.ok:
        click.secho(response.url)
        click.secho(response.text, fg="red")
        raise click.ClickException("Network request failed")
    return response


# def version_to_production(version: str) -> list:  # TODO make version_to_production
#    """Find the production name associated with a particular AnalysisProductions tag."""


def production_to_versions(working_group: str, production_name: str) -> list:
    """List the AnalysisProductions tags corresponding to the specified production."""
    datasets = apd.AnalysisData(working_group, production_name)
    versions = list(datasets._available_tags["version"])
    versions.sort()
    return versions



from DecayTreeTuple.Configuration import *
from Gaudi.Configuration import *
from DaVinci.Configuration import *
from PhysConf.Filters import LoKi_Filters
from Configurables import SubstitutePID
from Configurables import DecayTreeTuple, DaVinci,  LoKi__Hybrid__TupleTool, TupleToolDecayTreeFitter, TupleToolTrackIsolation, TupleToolTrigger, TupleToolRecoStats, TupleToolKinematic, TupleToolGeometry,TupleToolVtxIsoln, TupleToolConeIsolation


def addNtupleToDaVinci(isSim, stream, line):

    dtt = DecayTreeTuple('SignalTuple')
    dtt.Decay = "[ (B_s0 -> ^(D_s- -> ^K+ ^K- ^pi-) ^(tau+ -> ^pi+ ^pi- ^pi+ ) ) ]CC"

    dtt.ReFitPVs = True
    dtt.RevertToPositiveID = False

    dtt.addTupleTool("TupleToolPrimaries")
    #dtt.addTupleTool('TupleToolEventInfo')#already included
    dtt.addTupleTool("TupleToolVtxIsoln")

    #dtt.addTool(TupleToolTrackIsolation, name="TupleToolTrackIsolation")
    #dtt.TupleToolTrackIsolation.FillAsymmetry = False ##in philip era True che significa?????
    #dtt.TupleToolTrackIsolation.FillDeltaAngles = False
    #dtt.TupleToolTrackIsolation.MinConeAngle = 1.0

    dtt.addTool(TupleToolConeIsolation, name="TupleToolConeIsolation")
    dtt.TupleToolConeIsolation.ExtraParticlesLocation = 'Phys/StdAllNoPIDsPions/Particles'
    dtt.TupleToolConeIsolation.FillPi0Info = True
    dtt.TupleToolConeIsolation.MaxConeSize = 1.0


    #dtt.addTupleTool("TupleToolGeometry") #already included
    dtt.addTool( TupleToolGeometry, name = "TupleToolGeometry" )
    #dtt.TupleToolGeometry.OutputLevel = DEBUG
    dtt.TupleToolGeometry.Verbose = True
    dtt.TupleToolGeometry.RefitPVs = True
    dtt.TupleToolGeometry.FillMultiPV = True
    
    dtt.addTupleTool("TupleToolRecoStats")
    dtt.addTool(TupleToolRecoStats, name="TupleToolRecoStats")
    dtt.TupleToolRecoStats.Verbose = False

    dtt.addTupleTool("TupleToolTrackInfo") # ghostProb


    dtt.addBranches ({
        "B" : "^([ (B_s0 -> (D_s- -> K+ K- pi-) (tau+ -> pi+ pi- pi+  ) ) ]CC)",
        "Y" : "[ (B_s0 -> (D_s- -> K+ K- pi-) ^(tau+ -> pi+ pi- pi+  ) ) ]CC",
        "p1_fromY" : "[ (B_s0 -> (D_s- -> K+ K- pi-) (tau+ -> ^pi+ pi- pi+  ) ) ]CC",
        "p2_fromY" : "[ (B_s0 -> (D_s- -> K+ K- pi-) (tau+ -> pi+ ^pi- pi+  ) ) ]CC",
        "p3_fromY" : "[ (B_s0 -> (D_s- -> K+ K- pi-) (tau+ -> pi+ pi- ^pi+  ) ) ]CC",
        "Xc" : "[ (B_s0 -> ^(D_s- -> K+ K- pi-) (tau+ -> pi+ pi- pi+  ) ) ]CC",
        "p1_fromXc" : "[ (B_s0 -> (D_s- -> ^K+ K- pi-) (tau+ -> pi+ pi- pi+  ) ) ]CC",
        "p2_fromXc" : "[ (B_s0 -> (D_s- -> K+ ^K- pi-) (tau+ -> pi+ pi- pi+  ) ) ]CC",
        "p3_fromXc" : "[ (B_s0 -> (D_s- -> K+ K- ^pi-) (tau+ -> pi+ pi- pi+  ) ) ]CC"
    })

    tlist=["L0HadronDecision", "L0GlobalDecision",
           "Hlt1GlobalDecision","Hlt1TrackAllL0Decision","Hlt1TrackAllL0TightDecision",      #### AGGIUNTO
           ####"Hlt1TrackMVADecision","Hlt1TwoTrackMVADecision", # Non c'e' nel Run1
           "Hlt2GlobalDecision",
           "Hlt2Topo2BodyBBDTDecision","Hlt2Topo3BodyBBDTDecision" ,"Hlt2Topo4BodyBBDTDecision" ,
           "Hlt2Topo2BodySimpleDecision","Hlt2Topo3BodySimpleDecision","Hlt2Topo4BodySimpleDecision",
           "Hlt2IncPhiDecision","Hlt2IncPhiSidebandsDecision",
           "Hlt2CharmHadD2HHHDecision", "Hlt2CharmHadD2HHHWideMassDecision",  ### Non c'e'
           ]
    ########### ADDITIONAL Trigger info
    # 1. To get information about the TCK
    # dtt.addTupleTool("TupleToolEventInfo")
    # 2. To get information about trigger decisions
    ttt = dtt.addTupleTool("TupleToolTrigger")
    ttt.Verbose = True #Needed to get trigger decisions
    ttt.TriggerList = tlist

    #######################
    dtt.addTool(TupleToolDecay, name="B")

    #dttBt = dtt.B.addTool(TupleToolTrigger)
    #dttBt.TupleToolTrigger.Verbose = True
    #dttBt.TupleToolTrigger.TriggerList = tlist

    dttB = dtt.B.addTupleTool('TupleToolTISTOS')
    dttB.VerboseL0=True
    dttB.VerboseHlt1=True
    dttB.VerboseHlt2=True
    dttB.TriggerList = tlist

    from Configurables import MCMatchObjP2MCRelator
    
    if (isSim):
        dtt.ToolList +=  [
            "TupleToolMCBackgroundInfo",
            #"TupleToolPhotonInfo"
            ]
        
        # Add TupleToolMCTruth with fix for "Fatal error No valid data at '/Event/Hlt2/Long/Protos'"
        #default_rel_locs = MCMatchObjP2MCRelator().getDefaultProperty('RelTableLocations')
        #rel_locs = [loc for loc in default_rel_locs if 'Turbo' not in loc]
        
        mctruth = dtt.addTupleTool('TupleToolMCTruth')
        mctruth.ToolList =  [
            "MCTupleToolHierarchy",
	    "MCTupleToolKinematic"
            ]
        #mctruth.addTool(MCMatchObjP2MCRelator)
        #mctruth.MCMatchObjP2MCRelator.RelTableLocations = rel_locs


    LoKiTool= dtt.addTupleTool("LoKi::Hybrid::TupleTool/LoKiTool")
    LoKiTool.Variables = { "ETA" : "ETA" }

    dtt.B.addTupleTool('TupleToolDecayTreeFitter/DTF_Kpipi')
    dtt.B.DTF_Kpipi.constrainToOriginVertex = False
    dtt.B.DTF_Kpipi.Verbose = False
    dtt.B.DTF_Kpipi.Substitutions = {
    'Beauty -> Charm (tau+ -> ^pi+ pi- pi+ ) ': 'K+',
    'Beauty -> Charm (tau- -> ^pi- pi+ pi- ) ': 'K-',
    'Beauty -> X+ ^(Charm & X-)': 'D_s-',
    'Beauty -> X- ^(Charm & X+)': 'D_s+'}
    dtt.B.DTF_Kpipi.daughtersToConstrain = ["D_s-", "D_s+" ]
    dtt.B.DTF_Kpipi.UpdateDaughters = False

    dtt.B.addTupleTool('TupleToolDecayTreeFitter/DTF_piKpi')
    dtt.B.DTF_piKpi.constrainToOriginVertex = False
    dtt.B.DTF_piKpi.Verbose = False
    dtt.B.DTF_piKpi.Substitutions = {
    'Beauty -> Charm (tau+ -> pi+ ^pi- pi+ ) ': 'K-',
    'Beauty -> Charm (tau- -> pi- ^pi+ pi- ) ': 'K+',
    'Beauty -> X+ ^(Charm & X-)': 'D_s-',
    'Beauty -> X- ^(Charm & X+)': 'D_s+'}
    dtt.B.DTF_piKpi.daughtersToConstrain = ["D_s-", "D_s+" ]
    dtt.B.DTF_piKpi.UpdateDaughters = False

    dtt.B.addTupleTool('TupleToolDecayTreeFitter/DTF_pipiK')
    dtt.B.DTF_pipiK.constrainToOriginVertex = False
    dtt.B.DTF_pipiK.Verbose = False
    dtt.B.DTF_pipiK.Substitutions = {
    'Beauty -> Charm (tau+ -> pi+ pi- ^pi+ ) ': 'K+',
    'Beauty -> Charm (tau- -> pi- pi+ ^pi- ) ': 'K-',
    'Beauty -> X+ ^(Charm & X-)': 'D_s-',
    'Beauty -> X- ^(Charm & X+)': 'D_s+'}
    dtt.B.DTF_pipiK.daughtersToConstrain = ["D_s-", "D_s+" ]
    dtt.B.DTF_pipiK.UpdateDaughters = False



    dtt.B.addTool(TupleToolDecayTreeFitter("DTF"))
    dtt.B.ToolList +=  ["TupleToolDecayTreeFitter/DTF" ]
    dtt.B.DTF.constrainToOriginVertex = False
    dtt.B.DTF.Substitutions = { 'Beauty -> X+ ^(Charm & X-)' : 'D_s-', 'Beauty -> X- ^(Charm & X+)' : 'D_s+'  }
    dtt.B.DTF.daughtersToConstrain = ["D_s-", "D_s+", ]
    dtt.B.DTF.UpdateDaughters = True
    dtt.B.DTF.Verbose = True
    
    dtt.B.addTool(TupleToolDecayTreeFitter("DTFPVDs"))
    dtt.B.ToolList +=  ["TupleToolDecayTreeFitter/DTFPVDs" ]
    dtt.B.DTFPVDs.constrainToOriginVertex = True
    dtt.B.DTFPVDs.Substitutions = { 'Beauty -> X+ ^(Charm & X-)' : 'D_s-', 'Beauty -> X- ^(Charm & X+)' : 'D_s+' }
    dtt.B.DTFPVDs.daughtersToConstrain = ["D_s-", "D_s+" ]
    dtt.B.DTFPVDs.UpdateDaughters = True
    dtt.B.DTFPVDs.Verbose = True


    LoKiToolB=dtt.B.addTupleTool("LoKi::Hybrid::TupleTool/LoKiToolB")
    LoKiToolB.Variables = {
        "DOCA1" : "DOCA(1,2)",
        "TAU" : "BPVLTIME()",
        "TAUERR" : "BPVLTERR()" }

    dtt.addTool(TupleToolDecay, name="Xc")
    LoKiToolXc = dtt.Xc.addTupleTool("LoKi::Hybrid::TupleTool/LoKiToolXc")
    LoKiToolXc.Variables = { "DOCA1" : "DOCA(1,2)" , "DOCA2" : "DOCA(1,3)" , "DOCA3" : "DOCA(2,3)"}

    dtt.addTool(TupleToolDecay, name="Y")
    LoKiToolY = dtt.Y.addTupleTool("LoKi::Hybrid::TupleTool/LoKiToolY")
    LoKiToolY.Variables = { "DOCA1" : "DOCA(1,2)" , "DOCA2" : "DOCA(1,3)" , "DOCA3" : "DOCA(2,3)" }

    #dttXct = dtt.Xc.addTool(TupleToolTrigger)
    #dttXct.TupleToolTrigger.Verbose = True
    #dttXct.TupleToolTrigger.TriggerList = tlist

    dttXc = dtt.Xc.addTupleTool('TupleToolTISTOS')
    dttXc.VerboseL0=True
    dttXc.VerboseHlt1=True
    dttXc.VerboseHlt2=True
    dttXc.TriggerList= tlist

    
    # For 2018 data the line is explained here:
    # http://lhcbdoc.web.cern.ch/lhcbdoc/stripping/config/stripping34/bhadroncompleteevent/strippingbs2dstaunuforb2xtaunualllines.html
    # if isSim:
    #     stream = 'RXcHad.Strip'
    #     #stream = 'AllStreams'  #### MC di fondo non filtrato
    #     #line = 'Bs2DsTauNuForB2XTauNu'   # segnale
    #     #line = 'Bs2DsTauNuForB2XTauNuAllLines'   # segnale    
    #     line = line_name
    #     dtt.Inputs = ['/Event/{0}/Phys/{1}/Particles'.format(stream, line)]
    # else:
    #     stream = 'BhadronCompleteEvent'
    #     #stream = 'RXcHad.Strip'
    #     line = 'Bs2DsTauNuForB2XTauNu'
    #     dtt.Inputs = ['/Event/{0}/Phys/{1}/Particles'.format(stream,line)]

    # Setting the input based on the parameters passed
    dtt.Inputs = ['/Event/{0}/Phys/{1}/Particles'.format(stream,line)]

    # Configure DaVinci
    DaVinci().UserAlgorithms += [dtt]




###############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from collections import Counter
from io import StringIO

import click
import consolemd
from LbAPCommon.validators import explain_log
from LbAPCommon.validators.logs import _split_by_level


def show_log_advice(log_text, suppress=5):
    log_messages = _split_by_level(log_text)

    _show_message_summary(log_messages, "FATAL", fg="red", bold=True)
    _show_message_summary(log_messages, "ERROR", fg="red")
    _show_message_summary(log_messages, "WARNING", min_count=5, fg="yellow")
    click.echo()

    explanations, suggestions, errors = explain_log(log_text)

    if errors:
        click.secho("Errors have been detected!", fg="red", bold=True)
        _show_advice(errors)

    if suggestions:
        click.secho("Advice about warnings and errors:", fg="yellow")
        _show_advice(suggestions)

    if explanations:
        click.secho("General explanations", fg="green")
        _show_advice(explanations)

    if log_messages["FATAL"] or errors:
        raise click.ClickException("Found issues in log")


def _show_message_summary(log_messages, level, *, min_count=0, **formatting):
    total_count = len(log_messages[level])
    if not total_count:
        return
    messages = sorted(
        Counter(message for line_no, _, _, _, message in log_messages[level]).items(),
        key=lambda x: x[1],
        reverse=True,
    )
    click.secho(f"    Found {total_count} {level} message(s)", **formatting)
    n_supressed = 0
    n_supressed_unique = 0
    for message, count in messages:
        if count >= min_count:
            click.echo(f'        * {count} instances of "{message}"')
        else:
            n_supressed_unique += 1
            n_supressed += count
    if n_supressed:
        click.echo(
            f"        and {n_supressed} other(s) ({n_supressed_unique} unique). "
            'Pass "--suppress=0" to show all messages'
        )


def _show_advice(advice):
    for message, lines in advice:
        if lines is None:
            heading = "  * No line information available"
        else:
            heading = "  * Line" + ("s" if len(lines) > 1 else "") + ":"
            heading += f" {', '.join(map(str, lines[:5]))}"
            if len(lines) > 5:
                heading += f" and {len(lines) - 5} other(s)"
        click.secho(heading, fg="cyan")

        rendered_message = StringIO()
        consolemd.Renderer().render(message, output=rendered_message)
        click.echo("    " + rendered_message.getvalue().replace("\n", "\n    "))
